#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include "verify_main.h"
#include <fstream>
#include <sstream>
#include "nvbuf_utils.h"

SplicePipeManager pipeman;
static GstClockTime irscale_timestamp = 0;
static int is_dataupdate = 0;
unsigned char *testbuffer = NULL;
static pthread_mutex_t data_mutex = PTHREAD_MUTEX_INITIALIZER;
#define IR_YUV_DATA_SIZE (640*512*3/2)


static gboolean bus_call (GstBus * bus, GstMessage * msg, gpointer data)
{
    SplicePipeManager *app = (SplicePipeManager *)data;
    switch (GST_MESSAGE_TYPE (msg))
    {
        case GST_MESSAGE_ERROR:
        {
            GError *err = NULL;
            gchar *name, *debug = NULL;

            name = gst_object_get_path_string (msg->src);
            gst_message_parse_error (msg, &err, &debug);

            g_printerr ("ERROR on bus: by %s: %s\n", name, err->message);
            if (debug != NULL)
                g_printerr ("debug info:\n%s\n", debug);

            g_error_free (err);
            g_free (debug);
            g_free (name);

            //g_main_loop_quit (loop);
        }
        break;

        case GST_MESSAGE_STATE_CHANGED:
        {
            GstState old, new_state, pending;

            gst_message_parse_state_changed (msg, &old, &new_state, &pending);

            GST_DEBUG_OBJECT (GST_OBJECT (msg->src),
                              "changed state from %s to %s, pending %s\n",
                              gst_element_state_get_name (old), gst_element_state_get_name (new_state),
                              gst_element_state_get_name (pending));
            if (GST_MESSAGE_SRC (msg) == GST_OBJECT (app->irsplice_datasource_pipeline)
                && pending == GST_STATE_VOID_PENDING && old == GST_STATE_PAUSED
                && new_state == GST_STATE_PLAYING)
            {
            }
        }
        break;

        case GST_MESSAGE_EOS:
            g_print("End of Stream reached. \n");
            break;

        case GST_MESSAGE_APPLICATION:
        {
            const GstStructure *s;
            s = gst_message_get_structure (msg);

            if (gst_structure_has_name (s, "NvGstAppInterrupt"))
            {
                g_print ("Terminating the camera pipeline ...\n");
                //g_main_loop_quit (loop);
            }
        }
        break;

        case GST_MESSAGE_ELEMENT:
            g_printerr("Unexpected message received! \n");
            break;

        default:
            break;
    }
    return TRUE;
}

static GstFlowReturn ir_splice_scale_sample(GstElement *sink, void *data)
{
    GstSample *sample;
    int ret;
    g_signal_emit_by_name (sink, "pull-sample", &sample);

    if (sample)
    {
        GstBuffer *gstbuffer = gst_sample_get_buffer(sample);
        if(gstbuffer)
        {
            GstMapInfo map;
            if (gst_buffer_map(gstbuffer, &map, GST_MAP_READ))
            {
                //printf("result data size:%ld\r\n",map.size);
            }
            gst_buffer_unmap(gstbuffer, &map);
        }
        else
        {
            g_print("error: get buffer from %s failed \n", GST_OBJECT_NAME(sink));
        }
        gst_sample_unref (sample);
        return GST_FLOW_OK;
    }

    return GST_FLOW_ERROR;
}


static void irsplice_scale_need_data (GstElement * appsrc, guint unused,  gpointer user_data)
{
    int needframe = 1;
    GstBuffer *buffer = NULL;
    guint size = IR_YUV_DATA_SIZE;
    GstFlowReturn ret;
    GstMapInfo map;
    int getdataflag = 0;
    static int ir_framedrop_counter = 0;
    while(needframe)
    { 

        pthread_mutex_lock(&data_mutex);
        if(1 == is_dataupdate)
        {
            buffer = gst_buffer_new_allocate (NULL, size, NULL);
            if(NULL != buffer)
            {
                gst_buffer_map (buffer, &map, GST_MAP_WRITE);   
                memcpy(map.data,testbuffer, size);
                getdataflag = 1;
            }
            is_dataupdate = 0;
            pthread_mutex_unlock(&data_mutex);
            break;
        }
        else
        {
            pthread_mutex_unlock(&data_mutex);
            usleep(100);
        }
    }
    
    if(1 == getdataflag)
    {  
        //usleep(6*1000);
        GST_BUFFER_PTS (buffer) = irscale_timestamp;
        GST_BUFFER_DURATION (buffer) = gst_util_uint64_scale_int (1, GST_SECOND, 30);
        irscale_timestamp += GST_BUFFER_DURATION (buffer);
        gst_buffer_unmap(buffer, &map); 
        g_signal_emit_by_name (appsrc, "push-buffer", buffer, &ret);
        gst_buffer_unref (buffer);   
    }
}


int create_ir_splice_datasrc_pipeline()
{
    GstBus *bus = NULL;
    GstCapsFeatures *feature = NULL;
    GstCaps *caps = NULL;
    gboolean ret = TRUE;
    char crop_position[20] = {0};
    pipeman.irsplice_datasource_pipeline = gst_pipeline_new ("ir_splice_datasource_pipeline");
    if (!pipeman.irsplice_datasource_pipeline)
    {
        g_error ("irsplice_datasource_pipeline pipeline creation failed \n");
        goto fail;
    }
    bus = gst_pipeline_get_bus (GST_PIPELINE (pipeman.irsplice_datasource_pipeline));
    gst_bus_add_watch (bus, bus_call, (gpointer)&pipeman);
    gst_object_unref (bus);
    g_object_set (pipeman.irsplice_datasource_pipeline, "message-forward", TRUE, NULL);


    pipeman.irsplice_appsrc = gst_element_factory_make ("appsrc","irsplic-yuvappsrc");
    if (!pipeman.irsplice_appsrc)
    {
        g_error ("irappsrc  creation failed \n");
        goto fail;
    }
    gst_util_set_object_arg (G_OBJECT (pipeman.irsplice_appsrc), "format", "time");    
    g_object_set (G_OBJECT (pipeman.irsplice_appsrc), "caps",
                                                       gst_caps_new_simple ("video/x-raw", 
                                                       "format", G_TYPE_STRING, "NV12",
                                                       "width", G_TYPE_INT, 640,
                                                       "height", G_TYPE_INT, 512, 
                                                       //"blocksize",G_TYPE_INT,640*512*3/2,
                                                       "framerate",GST_TYPE_FRACTION, 30, 1, NULL), NULL);
    g_object_set (pipeman.irsplice_appsrc, "stream-type",  0, NULL);
    g_object_set (pipeman.irsplice_appsrc, "is-live", TRUE, NULL);
    g_object_set (pipeman.irsplice_appsrc, "min-percent",  0, NULL); 
    g_object_set (pipeman.irsplice_appsrc, "min-latency",  0, NULL);// then the min-latency must be set to 0 because it timestamps based on the running-time when the buffer entered appsrc
    g_object_set (pipeman.irsplice_appsrc, "max-latency",  0, NULL);    
    g_signal_connect (pipeman.irsplice_appsrc, "need-data", (GCallback) irsplice_scale_need_data, NULL);

    pipeman.irsplice_zoom_conv = gst_element_factory_make ("nvvideoconvert", "ir-yuvzoom-convert"); // NVGST_DEFAULT_VIDEO_CONVERTER_CSI
    if (!pipeman.irsplice_zoom_conv) 
    {
        g_error ("Element irsplice_zoom_conv creation failed \n");
        goto fail;
    }

    pipeman.irsplice_zoom_conv_out_filter = gst_element_factory_make ("capsfilter", NULL);
    if (!pipeman.irsplice_zoom_conv_out_filter) 
    {
        g_error ("svc_img_bin Element irsplice_zoom_conv_out_filterout filter creation failed \n");
        goto fail;
    }
    caps = gst_caps_new_simple ("video/x-raw", "format", G_TYPE_STRING, "NV12",
                                "width", G_TYPE_INT, 960,
                                "height", G_TYPE_INT, 768, NULL);
 
    g_object_set (pipeman.irsplice_zoom_conv_out_filter, "caps", caps, NULL);
    gst_caps_unref (caps);
    feature = NULL;
    caps = NULL;



    pipeman.irsplice_vidconv = gst_element_factory_make ("nvvideoconvert", "ir-yuv-convert"); // NVGST_DEFAULT_VIDEO_CONVERTER_CSI
    if (!pipeman.irsplice_vidconv) 
    {
        g_error ("Element video-convert creation failed \n");
        goto fail;
    }


    pipeman.irsplice_vidconv_out_filter = gst_element_factory_make ("capsfilter", NULL);
    if (!pipeman.irsplice_vidconv_out_filter) 
    {
        g_error ("svc_img_bin Element videoconv out filter creation failed \n");
        goto fail;
    }
    caps = gst_caps_new_simple ("video/x-raw", "format", G_TYPE_STRING, "NV12",
                                "width", G_TYPE_INT, 960,
                                "height", G_TYPE_INT, 1080, NULL);


    g_object_set (pipeman.irsplice_vidconv_out_filter, "caps", caps, NULL);
    gst_caps_unref (caps);
    feature = NULL;
    caps = NULL;


    sprintf(crop_position,"%d:%d:%d:%d",0,0,960,768);   //left:top:width:height    
    g_object_set (pipeman.irsplice_vidconv, "src-crop", crop_position, NULL);
    memset(crop_position,0,sizeof(crop_position));
    sprintf(crop_position,"%d:%d:%d:%d",0,156,960,768);
    g_object_set (pipeman.irsplice_vidconv, "dest-crop", crop_position, NULL);


    
    pipeman.irsplice_appsink = gst_element_factory_make ("appsink", "irspliceappsink");
    if (!pipeman.irsplice_appsink)
    {
        g_error ("irsplice_appsink  element could not be created.\n");
        goto fail;
    }

    g_object_set (pipeman.irsplice_zoom_conv, "compute-hw", 1, NULL);
    g_object_set (pipeman.irsplice_vidconv, "compute-hw", 1, NULL);


    gst_bin_add_many (GST_BIN ( pipeman.irsplice_datasource_pipeline), pipeman.irsplice_appsrc,pipeman.irsplice_zoom_conv, pipeman.irsplice_zoom_conv_out_filter,  pipeman.irsplice_vidconv, pipeman.irsplice_vidconv_out_filter, pipeman.irsplice_appsink,NULL);
    gst_element_link_many (pipeman.irsplice_appsrc, pipeman.irsplice_zoom_conv, pipeman.irsplice_zoom_conv_out_filter, pipeman.irsplice_vidconv, pipeman.irsplice_vidconv_out_filter,pipeman.irsplice_appsink,NULL);
    g_object_set(G_OBJECT(pipeman.irsplice_appsink), "emit-signals", TRUE, "sync", FALSE, NULL);
    g_signal_connect (pipeman.irsplice_appsink, "new-sample", G_CALLBACK(ir_splice_scale_sample), NULL);
    return TRUE;

fail:
    return FALSE;
}
 
int run_pipeline()
{
    if (GST_STATE_CHANGE_FAILURE == gst_element_set_state (pipeman.irsplice_datasource_pipeline, GST_STATE_PLAYING))
    {
        printf ("can't set ir-splice pipeline to playing\n");
        return false;
    }
    printf("set pipeline start success!\r\n");
    return true;
}

void* simulate_feed_data_into_the_pipeline(void *idx)
{
    int tempvalue = 0;
    while(1)
    {
        //In practice, this will generate real-time data queue for transport to appsrc,This is just a simulation
        tempvalue++;
        if(15 == tempvalue)
        {
            tempvalue=0;
        }
        pthread_mutex_lock(&data_mutex);
        is_dataupdate = 1;
        memset(testbuffer,tempvalue,sizeof(testbuffer));
        pthread_mutex_unlock(&data_mutex);
        usleep(1000/30*1000);
    }
    return NULL;
}


int main(int argc, char **argv)
{
    testbuffer = (unsigned char*)malloc(IR_YUV_DATA_SIZE*sizeof(unsigned char));
    if(testbuffer == NULL)
    {
        printf("testbuffer malloc fail!\r\n");
        return -1;
    }

    pthread_t t;
    gst_init (NULL, NULL);
  
    create_ir_splice_datasrc_pipeline();
    run_pipeline();
    pthread_create(&t, NULL, simulate_feed_data_into_the_pipeline, NULL);
    sleep(1);

    while(1)
    {
        sleep(5);
    }
    return 0;
}